@@ -1,7 +1,7 @@
 <?php
 /*
 ** Zabbix
-** Copyright (C) 2001-2020 Zabbix SIA
+** Copyright (C) 2001-2021 Zabbix SIA
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
@@ -128,6 +128,20 @@ abstract class CItemGeneral extends CApiService {
 	 * @param bool  $update
 	 */
 	protected function checkInput(array &$items, $update = false) {
+		$api_input_rules = ['type' => API_OBJECT, 'fields' => [
+			'type' => ['type' => API_INT32, 'flags' => API_REQUIRED, 'in' => implode(',', static::SUPPORTED_ITEM_TYPES)]
+		]];
+		if ($update) {
+			unset($api_input_rules['fields']['type']['flags']);
+		}
+
+		foreach ($items as $num => $item) {
+			$data = array_intersect_key($item, $api_input_rules['fields']);
+			if (!CApiInputValidator::validate($api_input_rules, $data, '/'.($num + 1), $error)) {
+				self::exception(ZBX_API_ERROR_PARAMETERS, $error);
+			}
+		}
+
 		if ($update) {
 			$itemDbFields = ['itemid' => null];
 
@@ -150,7 +164,6 @@ abstract class CItemGeneral extends CApiService {
 				'hostids' => zbx_objectValues($dbItems, 'hostid'),
 				'templated_hosts' => true,
 				'editable' => true,
-				'selectApplications' => ['applicationid', 'flags'],
 				'preservekeys' => true
 			]);
 		}
@@ -169,7 +182,6 @@ abstract class CItemGeneral extends CApiService {
 				'hostids' => zbx_objectValues($items, 'hostid'),
 				'templated_hosts' => true,
 				'editable' => true,
-				'selectApplications' => ['applicationid', 'flags'],
 				'preservekeys' => true
 			]);
 
@@ -249,8 +261,10 @@ abstract class CItemGeneral extends CApiService {
 			'lldmacros' => (get_class($this) === 'CItemPrototype')
 		]);
 
+		$index = 0;
 		foreach ($items as $inum => &$item) {
 			$item = $this->clearValues($item);
+			$index++;
 
 			$fullItem = $items[$inum];
 
@@ -334,7 +348,8 @@ abstract class CItemGeneral extends CApiService {
 
 			if ($fullItem['type'] == ITEM_TYPE_CALCULATED) {
 				$api_input_rules = ['type' => API_OBJECT, 'fields' => [
-					'params' => ['type' => API_CALC_FORMULA, 'flags' => $this instanceof CItemPrototype ? API_ALLOW_LLD_MACRO : 0]
+					'params' =>		['type' => API_CALC_FORMULA, 'flags' => $this instanceof CItemPrototype ? API_ALLOW_LLD_MACRO : 0, 'length' => DB::getFieldLength('items', 'params')],
+					'value_type' =>	['type' => API_INT32, 'in' => ITEM_VALUE_TYPE_UINT64.','.ITEM_VALUE_TYPE_FLOAT]
 				]];
 
 				$data = array_intersect_key($item, $api_input_rules['fields']);
@@ -448,27 +463,6 @@ abstract class CItemGeneral extends CApiService {
 				);
 			}
 
-			// parameters
-			if ($fullItem['type'] == ITEM_TYPE_AGGREGATE) {
-				$params_num = $item_key_parser->getParamsNum();
-
-				if (!str_in_array($item_key_parser->getKey(), ['grpmax', 'grpmin', 'grpsum', 'grpavg'])
-						|| $params_num > 4 || $params_num < 3
-						|| ($params_num == 3 && $item_key_parser->getParam(2) !== 'last')
-						|| !str_in_array($item_key_parser->getParam(2), ['last', 'min', 'max', 'avg', 'sum', 'count'])) {
-					self::exception(ZBX_API_ERROR_PARAMETERS,
-						_s('Key "%1$s" does not match <grpmax|grpmin|grpsum|grpavg>["Host group(s)", "Item key",'.
-							' "<last|min|max|avg|sum|count>", "parameter"].', $item_key_parser->getKey()));
-				}
-			}
-
-			// type of information
-			if ($fullItem['type'] == ITEM_TYPE_AGGREGATE && $fullItem['value_type'] != ITEM_VALUE_TYPE_UINT64
-					&& $fullItem['value_type'] != ITEM_VALUE_TYPE_FLOAT) {
-					self::exception(ZBX_API_ERROR_PARAMETERS,
-						_('Type of information must be "Numeric (unsigned)" or "Numeric (float)" for aggregate items.'));
-			}
-
 			if (($fullItem['type'] == ITEM_TYPE_TRAPPER || $fullItem['type'] == ITEM_TYPE_HTTPAGENT)
 					&& array_key_exists('trapper_hosts', $fullItem) && $fullItem['trapper_hosts'] !== ''
 					&& !$ip_range_parser->parse($fullItem['trapper_hosts'])) {
@@ -578,38 +572,98 @@ abstract class CItemGeneral extends CApiService {
 				self::exception(ZBX_API_ERROR_PARAMETERS, _('No SNMP OID specified.'));
 			}
 
-			if (isset($item['applications']) && $item['applications']) {
-				/*
-				 * 'flags' is available for update and item prototypes.
-				 * Don't allow discovered or any other application types for item prototypes in 'applications' option.
-				 */
-				if (array_key_exists('flags', $fullItem) && $fullItem['flags'] == ZBX_FLAG_DISCOVERY_PROTOTYPE) {
-					foreach ($host['applications'] as $num => $application) {
-						if ($application['flags'] != ZBX_FLAG_DISCOVERY_NORMAL) {
-							unset($host['applications'][$num]);
-						}
-					}
-				}
-
-				// check that the given applications belong to the item's host
-				$dbApplicationIds = zbx_objectValues($host['applications'], 'applicationid');
-				foreach ($item['applications'] as $appId) {
-					if (!in_array($appId, $dbApplicationIds)) {
-						$error = _s('Application with ID "%1$s" is not available on "%2$s".', $appId, $host['name']);
-						self::exception(ZBX_API_ERROR_PARAMETERS, $error);
-					}
-				}
-			}
-
 			$this->checkSpecificFields($fullItem, $update ? 'update' : 'create');
 
 			$this->validateItemPreprocessing($fullItem);
+			$this->validateTags($item, '/'.$index);
 		}
 		unset($item);
 
+		$this->validateValueMaps($items);
+
+		$this->checkAndAddUuid($items, $dbHosts, $update);
 		$this->checkExistingItems($items);
 	}
 
+	/**
+	 * Check that only items on templates have UUID. Add UUID to all host prototypes on templates,
+	 *   if it doesn't exist.
+	 *
+	 * @param array $items_to_create
+	 * @param array $db_hosts
+	 * @param bool $is_update
+	 *
+	 * @throws APIException
+	 */
+	protected function checkAndAddUuid(array &$items_to_create, array $db_hosts, bool $is_update): void {
+		if ($is_update) {
+			foreach ($items_to_create as $index => &$item) {
+				if (array_key_exists('uuid', $item)) {
+					self::exception(ZBX_API_ERROR_PARAMETERS,
+						_s('Invalid parameter "%1$s": %2$s.', '/' . ($index + 1),
+							_s('unexpected parameter "%1$s"', 'uuid')
+						)
+					);
+				}
+			}
+
+			return;
+		}
+
+		foreach ($items_to_create as $index => &$item) {
+			if ($db_hosts[$item['hostid']]['status'] != HOST_STATUS_TEMPLATE && array_key_exists('uuid', $item)) {
+				self::exception(ZBX_API_ERROR_PARAMETERS,
+					_s('Invalid parameter "%1$s": %2$s.', '/' . ($index + 1), _s('unexpected parameter "%1$s"', 'uuid'))
+				);
+			}
+
+			if ($db_hosts[$item['hostid']]['status'] == HOST_STATUS_TEMPLATE && !array_key_exists('uuid', $item)) {
+				$item['uuid'] = generateUuidV4();
+			}
+		}
+		unset($item);
+
+		$db_uuid = DB::select('items', [
+			'output' => ['uuid'],
+			'filter' => ['uuid' => array_column($items_to_create, 'uuid')],
+			'limit' => 1
+		]);
+
+		if ($db_uuid) {
+			self::exception(ZBX_API_ERROR_PARAMETERS,
+				_s('Entry with UUID "%1$s" already exists.', $db_uuid[0]['uuid'])
+			);
+		}
+	}
+
+	/**
+	 * Validates tags.
+	 *
+	 * @param array  $item
+	 * @param array  $item['tags']
+	 * @param string $item['tags'][]['tag']
+	 * @param string $item['tags'][]['value']
+	 *
+	 * @throws APIException if the input is invalid.
+	 */
+	protected function validateTags(array $item, string $path = '/') {
+		if (!array_key_exists('tags', $item)) {
+			return;
+		}
+
+		$api_input_rules = ['type' => API_OBJECT, 'fields' => [
+			'tags'		=> ['type' => API_OBJECTS, 'uniq' => [['tag', 'value']], 'fields' => [
+				'tag'		=> ['type' => API_STRING_UTF8, 'flags' => API_REQUIRED | API_NOT_EMPTY, 'length' => DB::getFieldLength('item_tag', 'tag')],
+				'value'		=> ['type' => API_STRING_UTF8, 'length' => DB::getFieldLength('item_tag', 'value')]
+			]]
+		]];
+
+		$item_tags = ['tags' => $item['tags']];
+		if (!CApiInputValidator::validate($api_input_rules, $item_tags, $path, $error)) {
+			self::exception(ZBX_API_ERROR_PARAMETERS, $error);
+		}
+	}
+
 	/**
 	 * Check item specific fields. Each API like Item, Itemprototype and Discovery rule may inherit different fields
 	 * to validate.
@@ -804,10 +858,6 @@ abstract class CItemGeneral extends CApiService {
 	 * @param string     $tpl_items[<itemid>]['hostid']
 	 * @param string     $tpl_items[<itemid>]['key_']
 	 * @param int        $tpl_items[<itemid>]['type']
-	 * @param array      $tpl_items[<itemid>]['applicationPrototypes']            (optional) Suitable for item
-	 *                                                                            prototypes.
-	 * @param string     $tpl_items[<itemid>]['applicationPrototypes'][]['name']
-	 * @param array      $tpl_items[<itemid>]['applications']                     (optional) Array of applicationids.
 	 * @param array      $tpl_items[<itemid>]['preprocessing']                    (optional)
 	 * @param int        $tpl_items[<itemid>]['preprocessing'][]['type']
 	 * @param string     $tpl_items[<itemid>]['preprocessing'][]['params']
@@ -898,32 +948,6 @@ abstract class CItemGeneral extends CApiService {
 			}
 		}
 
-		// Preparing list of application prototypes.
-		if ($this instanceof CItemPrototype) {
-			$tpl_app_prototypes = [];
-			$item_prototypeids = [];
-
-			foreach ($tpl_items as $tpl_item) {
-				if (array_key_exists('applicationPrototypes', $tpl_item) && $tpl_item['applicationPrototypes']) {
-					$item_prototypeids[] = $tpl_item['itemid'];
-				}
-			}
-
-			if ($item_prototypeids) {
-				$db_tpl_app_prototypes = DBselect(
-					'SELECT iap.itemid,iap.application_prototypeid,ap.name'.
-					' FROM item_application_prototype iap,application_prototype ap'.
-					' WHERE iap.application_prototypeid=ap.application_prototypeid'.
-						' AND '.dbConditionInt('iap.itemid', $item_prototypeids)
-				);
-
-				while ($db_tpl_app_prototype = DBfetch($db_tpl_app_prototypes)) {
-					$tpl_app_prototypes[$db_tpl_app_prototype['itemid']][$db_tpl_app_prototype['name']] =
-						$db_tpl_app_prototype['application_prototypeid'];
-				}
-			}
-		}
-
 		// List of the discovery rules.
 		if ($this instanceof CItemPrototype) {
 			// List of itemids without 'ruleid' property.
@@ -1024,6 +1048,8 @@ abstract class CItemGeneral extends CApiService {
 
 				// copying item
 				$new_item = $tpl_item;
+				$new_item['uuid'] = '';
+
 				if ($chd_item !== null) {
 					$new_item['itemid'] = $chd_item['itemid'];
 				}
@@ -1068,14 +1094,6 @@ abstract class CItemGeneral extends CApiService {
 					}
 				}
 
-				if ($this instanceof CItemPrototype && array_key_exists('applicationPrototypes', $new_item)) {
-					foreach ($new_item['applicationPrototypes'] as &$application_prototype) {
-						$application_prototype['templateid'] =
-							$tpl_app_prototypes[$tpl_item['itemid']][$application_prototype['name']];
-					}
-					unset($application_prototype);
-				}
-
 				$new_items[] = $new_item;
 			}
 		}
@@ -1099,46 +1117,6 @@ abstract class CItemGeneral extends CApiService {
 			}
 		}
 
-		// Setting item applications.
-		if ($this instanceof CItem || $this instanceof CItemPrototype) {
-			$tpl_applicationids = [];
-			foreach ($tpl_items as $tpl_item) {
-				if (array_key_exists('applications', $tpl_item)) {
-					foreach ($tpl_item['applications'] as $applicationid) {
-						$tpl_applicationids[$applicationid] = true;
-					}
-				}
-			}
-
-			if ($tpl_applicationids) {
-				$db_applications = DBselect('SELECT a.hostid,at.templateid,a.applicationid'.
-					' FROM application_template at,applications a'.
-					' WHERE at.applicationid=a.applicationid'.
-						' AND '.dbConditionInt('at.templateid', array_keys($tpl_applicationids)).
-						' AND '.dbConditionInt('a.hostid', array_keys($chd_hosts))
-				);
-
-				$app_links = [];
-				while ($db_application = DBfetch($db_applications)) {
-					$app_links[$db_application['hostid']][$db_application['templateid']] =
-						$db_application['applicationid'];
-				}
-
-				foreach ($new_items as &$new_item) {
-					if (array_key_exists('applications', $new_item)) {
-						$applicationids = [];
-						foreach ($new_item['applications'] as $applicationid) {
-							if (array_key_exists($applicationid, $app_links[$new_item['hostid']])) {
-								$applicationids[] =  $app_links[$new_item['hostid']][$applicationid];
-							}
-						}
-						$new_item['applications'] = $applicationids;
-					}
-				}
-				unset($new_item);
-			}
-		}
-
 		return $this->prepareDependentItems($tpl_items, $new_items, $hostids);
 	}
 
@@ -1244,7 +1222,7 @@ abstract class CItemGeneral extends CApiService {
 				self::exception(ZBX_API_ERROR_PARAMETERS, _('Incorrect arguments passed to function.'));
 			}
 
-			$type_validator = new CLimitedSetValidator(['values' => $this::$supported_preprocessing_types]);
+			$type_validator = new CLimitedSetValidator(['values' => static::SUPPORTED_PREPROCESSING_TYPES]);
 
 			$error_handler_validator = new CLimitedSetValidator([
 				'values' => [ZBX_PREPROC_FAIL_DEFAULT, ZBX_PREPROC_FAIL_DISCARD_VALUE, ZBX_PREPROC_FAIL_SET_VALUE,
@@ -1252,6 +1230,10 @@ abstract class CItemGeneral extends CApiService {
 				]
 			]);
 
+			$unsupported_error_handler_validator = new CLimitedSetValidator([
+				'values' => [ZBX_PREPROC_FAIL_DISCARD_VALUE, ZBX_PREPROC_FAIL_SET_VALUE, ZBX_PREPROC_FAIL_SET_ERROR]
+			]);
+
 			$prometheus_pattern_parser = new CPrometheusPatternParser(['usermacros' => true,
 				'lldmacros' => ($this instanceof CItemPrototype)
 			]);
@@ -1446,6 +1428,7 @@ abstract class CItemGeneral extends CApiService {
 
 					case ZBX_PREPROC_DELTA_VALUE:
 					case ZBX_PREPROC_DELTA_SPEED:
+					case ZBX_PREPROC_XML_TO_JSON:
 						// Check if 'params' is empty, because it must be empty.
 						if (is_array($preprocessing['params'])) {
 							self::exception(ZBX_API_ERROR_PARAMETERS, _('Incorrect arguments passed to function.'));
@@ -1457,12 +1440,15 @@ abstract class CItemGeneral extends CApiService {
 							);
 						}
 
-						// Check if one of the deltas (Delta per second or Delta value) already exists.
-						if ($delta) {
-							self::exception(ZBX_API_ERROR_PARAMETERS, _('Only one change step is allowed.'));
-						}
-						else {
-							$delta = true;
+						if ($preprocessing['type'] == ZBX_PREPROC_DELTA_VALUE
+								|| $preprocessing['type'] == ZBX_PREPROC_DELTA_SPEED) {
+							// Check if one of the deltas (Delta per second or Delta value) already exists.
+							if ($delta) {
+								self::exception(ZBX_API_ERROR_PARAMETERS, _('Only one change step is allowed.'));
+							}
+							else {
+								$delta = true;
+							}
 						}
 						break;
 
@@ -1652,6 +1638,43 @@ abstract class CItemGeneral extends CApiService {
 						}
 						break;
 
+					case ZBX_PREPROC_VALIDATE_NOT_SUPPORTED:
+						if (is_array($preprocessing['error_handler'])) {
+							self::exception(ZBX_API_ERROR_PARAMETERS, _('Incorrect arguments passed to function.'));
+						}
+						elseif (!$unsupported_error_handler_validator->validate($preprocessing['error_handler'])) {
+							self::exception(ZBX_API_ERROR_PARAMETERS,
+								_s('Incorrect value for field "%1$s": %2$s.', 'error_handler',
+									_s('unexpected value "%1$s"', $preprocessing['error_handler'])
+								)
+							);
+						}
+
+						if (is_array($preprocessing['error_handler_params'])) {
+							self::exception(ZBX_API_ERROR_PARAMETERS, _('Incorrect arguments passed to function.'));
+						}
+						elseif ($preprocessing['error_handler'] == ZBX_PREPROC_FAIL_DISCARD_VALUE
+								&& $preprocessing['error_handler_params'] !== ''
+								&& $preprocessing['error_handler_params'] !== null
+								&& $preprocessing['error_handler_params'] !== false) {
+							self::exception(ZBX_API_ERROR_PARAMETERS,
+								_s('Incorrect value for field "%1$s": %2$s.', 'error_handler_params',
+									_('should be empty')
+								)
+							);
+						}
+						elseif ($preprocessing['error_handler'] == ZBX_PREPROC_FAIL_SET_ERROR
+								&& ($preprocessing['error_handler_params'] === ''
+									|| $preprocessing['error_handler_params'] === null
+									|| $preprocessing['error_handler_params'] === false)) {
+							self::exception(ZBX_API_ERROR_PARAMETERS,
+								_s('Incorrect value for field "%1$s": %2$s.', 'error_handler_params',
+									_('cannot be empty')
+								)
+							);
+						}
+						break;
+
 					default:
 						if (is_array($preprocessing['error_handler'])) {
 							self::exception(ZBX_API_ERROR_PARAMETERS, _('Incorrect arguments passed to function.'));
@@ -1715,8 +1738,9 @@ abstract class CItemGeneral extends CApiService {
 	/**
 	 * Insert item pre-processing data into DB.
 	 *
-	 * @param array $items                     An array of items.
-	 * @param array $items[]['preprocessing']  An array of item pre-processing data.
+	 * @param array  $items                     An array of items.
+	 * @param string $items[]['itemid']
+	 * @param array  $items[]['preprocessing']  An array of item pre-processing data.
 	 */
 	protected function createItemPreprocessing(array $items) {
 		$item_preproc = [];
@@ -1728,7 +1752,7 @@ abstract class CItemGeneral extends CApiService {
 				foreach ($item['preprocessing'] as $preprocessing) {
 					$item_preproc[] = [
 						'itemid' => $item['itemid'],
-						'step' => $step++,
+						'step' => ($preprocessing['type'] == ZBX_PREPROC_VALIDATE_NOT_SUPPORTED) ? 0 : $step++,
 						'type' => $preprocessing['type'],
 						'params' => $preprocessing['params'],
 						'error_handler' => $preprocessing['error_handler'],
@@ -1746,37 +1770,98 @@ abstract class CItemGeneral extends CApiService {
 	/**
 	 * Update item pre-processing data in DB. Delete old records and create new ones.
 	 *
-	 * @param array $items                     An array of items.
-	 * @param array $items[]['preprocessing']  An array of item pre-processing data.
+	 * @param array  $items
+	 * @param string $items[]['itemid']
+	 * @param array  $items[]['preprocessing']
+	 * @param int    $items[]['preprocessing'][]['type']
+	 * @param string $items[]['preprocessing'][]['params']
+	 * @param int    $items[]['preprocessing'][]['error_handler']
+	 * @param string $items[]['preprocessing'][]['error_handler_params']
 	 */
 	protected function updateItemPreprocessing(array $items) {
-		$item_preproc = [];
-		$item_preprocids = [];
+		$item_preprocs = [];
 
 		foreach ($items as $item) {
 			if (array_key_exists('preprocessing', $item)) {
-				$item_preprocids[] = $item['itemid'];
+				$item_preprocs[$item['itemid']] = [];
 				$step = 1;
 
-				foreach ($item['preprocessing'] as $preprocessing) {
-					$item_preproc[] = [
-						'itemid' => $item['itemid'],
-						'step' => $step++,
-						'type' => $preprocessing['type'],
-						'params' => $preprocessing['params'],
-						'error_handler' => $preprocessing['error_handler'],
-						'error_handler_params' => $preprocessing['error_handler_params']
+				foreach ($item['preprocessing'] as $item_preproc) {
+					$curr_step = ($item_preproc['type'] == ZBX_PREPROC_VALIDATE_NOT_SUPPORTED) ? 0 : $step++;
+					$item_preprocs[$item['itemid']][$curr_step] = [
+						'type' => $item_preproc['type'],
+						'params' => $item_preproc['params'],
+						'error_handler' => $item_preproc['error_handler'],
+						'error_handler_params' => $item_preproc['error_handler_params']
 					];
 				}
 			}
 		}
 
-		if ($item_preprocids) {
-			DB::delete('item_preproc', ['itemid' => $item_preprocids]);
+		if (!$item_preprocs) {
+			return;
 		}
 
-		if ($item_preproc) {
-			DB::insertBatch('item_preproc', $item_preproc);
+		$ins_item_preprocs = [];
+		$upd_item_preprocs = [];
+		$del_item_preprocids = [];
+
+		$options = [
+			'output' => ['item_preprocid', 'itemid', 'step', 'type', 'params', 'error_handler', 'error_handler_params'],
+			'filter' => ['itemid' => array_keys($item_preprocs)]
+		];
+		$db_item_preprocs = DBselect(DB::makeSql('item_preproc', $options));
+
+		while ($db_item_preproc = DBfetch($db_item_preprocs)) {
+			if (array_key_exists($db_item_preproc['step'], $item_preprocs[$db_item_preproc['itemid']])) {
+				$item_preproc = $item_preprocs[$db_item_preproc['itemid']][$db_item_preproc['step']];
+				$upd_item_preproc = [];
+
+				if ($item_preproc['type'] != $db_item_preproc['type']) {
+					$upd_item_preproc['type'] = $item_preproc['type'];
+				}
+				if ($item_preproc['params'] !== $db_item_preproc['params']) {
+					$upd_item_preproc['params'] = $item_preproc['params'];
+				}
+				if ($item_preproc['error_handler'] != $db_item_preproc['error_handler']) {
+					$upd_item_preproc['error_handler'] = $item_preproc['error_handler'];
+				}
+				if ($item_preproc['error_handler_params'] !== $db_item_preproc['error_handler_params']) {
+					$upd_item_preproc['error_handler_params'] = $item_preproc['error_handler_params'];
+				}
+
+				if ($upd_item_preproc) {
+					$upd_item_preprocs[] = [
+						'values' => $upd_item_preproc,
+						'where' => ['item_preprocid' => $db_item_preproc['item_preprocid']]
+					];
+				}
+				unset($item_preprocs[$db_item_preproc['itemid']][$db_item_preproc['step']]);
+			}
+			else {
+				$del_item_preprocids[] = $db_item_preproc['item_preprocid'];
+			}
+		}
+
+		foreach ($item_preprocs as $itemid => $preprocs) {
+			foreach ($preprocs as $step => $preproc) {
+				$ins_item_preprocs[] = [
+					'itemid' => $itemid,
+					'step' => $step
+				] + $preproc;
+			}
+		}
+
+		if ($del_item_preprocids) {
+			DB::delete('item_preproc', ['item_preprocid' => $del_item_preprocids]);
+		}
+
+		if ($upd_item_preprocs) {
+			DB::update('item_preproc', $upd_item_preprocs);
+		}
+
+		if ($ins_item_preprocs) {
+			DB::insertBatch('item_preproc', $ins_item_preprocs);
 		}
 	}
 
@@ -1981,6 +2066,52 @@ abstract class CItemGeneral extends CApiService {
 			}
 		}
 
+		// Add value mapping.
+		if (($this instanceof CItemPrototype || $this instanceof CItem) && $options['selectValueMap'] !== null) {
+			if ($options['selectValueMap'] === API_OUTPUT_EXTEND) {
+				$options['selectValueMap'] = ['valuemapid', 'name', 'mappings'];
+			}
+
+			foreach ($result as &$item) {
+				$item['valuemap'] = [];
+			}
+			unset($item);
+
+			$valuemaps = DB::select('valuemap', [
+				'output' => array_diff($this->outputExtend($options['selectValueMap'], ['valuemapid', 'hostid']),
+					['mappings']
+				),
+				'filter' => ['valuemapid' => array_keys(array_flip(array_column($result, 'valuemapid')))],
+				'preservekeys' => true
+			]);
+
+			if ($this->outputIsRequested('mappings', $options['selectValueMap']) && $valuemaps) {
+				$params = [
+					'output' => ['valuemapid', 'type', 'value', 'newvalue'],
+					'filter' => ['valuemapid' => array_keys($valuemaps)],
+					'sortfield' => ['sortorder']
+				];
+				$query = DBselect(DB::makeSql('valuemap_mapping', $params));
+
+				while ($mapping = DBfetch($query)) {
+					$valuemaps[$mapping['valuemapid']]['mappings'][] = [
+						'type' => $mapping['type'],
+						'value' => $mapping['value'],
+						'newvalue' => $mapping['newvalue']
+					];
+				}
+			}
+
+			foreach ($result as &$item) {
+				if (array_key_exists('valuemapid', $item) && array_key_exists($item['valuemapid'], $valuemaps)) {
+					$item['valuemap'] = array_intersect_key($valuemaps[$item['valuemapid']],
+						array_flip($options['selectValueMap'])
+					);
+				}
+			}
+			unset($item);
+		}
+
 		if (!$options['countOutput'] && $this->outputIsRequested('parameters', $options['output'])) {
 			$item_parameters = DBselect(
 				'SELECT ip.itemid,ip.name,ip.value'.
@@ -2589,4 +2720,135 @@ abstract class CItemGeneral extends CApiService {
 
 		return $result;
 	}
+
+	/**
+	 * Update item tags.
+	 *
+	 * @param array  $items
+	 * @param string $items[]['itemid']
+	 * @param array  $items[]['tags']
+	 * @param string $items[]['tags'][]['tag']
+	 * @param string $items[]['tags'][]['value']
+	 */
+	protected function updateItemTags(array $items): void {
+		$items = array_filter($items, function ($item) {
+			return array_key_exists('tags', $item);
+		});
+
+		// Select tags from database.
+		$db_tags = DBselect(
+			'SELECT itemtagid, itemid, tag, value'.
+			' FROM item_tag'.
+			' WHERE '.dbConditionInt('itemid', array_keys($items))
+		);
+
+		array_walk($items, function (&$item) {
+			$item['db_tags'] = [];
+		});
+
+		while ($db_tag = DBfetch($db_tags)) {
+			$items[$db_tag['itemid']]['db_tags'][] = $db_tag;
+		}
+
+		// Find which tags must be added/deleted.
+		$new_tags = [];
+		$del_tagids = [];
+		foreach ($items as $item) {
+			CArrayHelper::sort($item['tags'], ['tag', 'value']);
+
+			foreach ($item['db_tags'] as $del_tag_key => $tag_delete) {
+				foreach ($item['tags'] as $new_tag_key => $tag_add) {
+					if ($tag_delete['tag'] === $tag_add['tag'] && $tag_delete['value'] === $tag_add['value']) {
+						unset($item['db_tags'][$del_tag_key], $item['tags'][$new_tag_key]);
+						continue 2;
+					}
+				}
+			}
+
+			$del_tagids = array_merge($del_tagids, array_column($item['db_tags'], 'itemtagid'));
+
+			foreach ($item['tags'] as $tag_add) {
+				$tag_add['itemid'] = $item['itemid'];
+				$new_tags[] = $tag_add;
+			}
+		}
+
+		if ($del_tagids) {
+			DB::delete('item_tag', ['itemtagid' => $del_tagids]);
+		}
+		if ($new_tags) {
+			DB::insert('item_tag', $new_tags);
+		}
+	}
+
+	/**
+	 * Record item tags into database.
+	 *
+	 * @param array  $items
+	 * @param array  $items[]['tags']
+	 * @param string $items[]['tags'][]['tag']
+	 * @param string $items[]['tags'][]['value']
+	 * @param int    $items[]['itemid']
+	 */
+	protected function createItemTags(array $items): void {
+		$new_tags = [];
+		foreach ($items as $key => $item) {
+			if (array_key_exists('tags', $item)) {
+				foreach ($item['tags'] as $tag) {
+					$tag['itemid'] = $item['itemid'];
+					$new_tags[] = $tag;
+				}
+			}
+		}
+
+		if ($new_tags) {
+			DB::insert('item_tag', $new_tags);
+		}
+	}
+
+	/**
+	 * Check that valuemap belong to same host as item.
+	 *
+	 * @param array $items
+	 */
+	protected function validateValueMaps(array $items): void {
+		$valuemapids_by_hostid = [];
+
+		foreach ($items as $item) {
+			if (array_key_exists('valuemapid', $item) && $item['valuemapid'] != 0) {
+				$valuemapids_by_hostid[$item['hostid']][$item['valuemapid']] = true;
+			}
+		}
+
+		$sql_where = [];
+		foreach ($valuemapids_by_hostid as $hostid => $valuemapids) {
+			$sql_where[] = '(vm.hostid='.zbx_dbstr($hostid).' AND '.
+				dbConditionId('vm.valuemapid', array_keys($valuemapids)).')';
+		}
+
+		if ($sql_where) {
+			$result = DBselect(
+				'SELECT vm.valuemapid,vm.hostid'.
+				' FROM valuemap vm'.
+				' WHERE '.implode(' OR ', $sql_where)
+			);
+			while ($row = DBfetch($result)) {
+				unset($valuemapids_by_hostid[$row['hostid']][$row['valuemapid']]);
+
+				if (!$valuemapids_by_hostid[$row['hostid']]) {
+					unset($valuemapids_by_hostid[$row['hostid']]);
+				}
+			}
+
+			if ($valuemapids_by_hostid) {
+				$hostid = key($valuemapids_by_hostid);
+				$valuemapid = key($valuemapids_by_hostid[$hostid]);
+
+				$host_row = DBfetch(DBselect('SELECT h.host FROM hosts h WHERE h.hostid='.zbx_dbstr($hostid)));
+				self::exception(ZBX_API_ERROR_PARAMETERS, _s('Valuemap with ID "%1$s" is not available on "%2$s".',
+					$valuemapid, $host_row['host']
+				));
+			}
+		}
+	}
 }
