@@ -1,7 +1,7 @@
 <?php
 /*
 ** Zabbix
-** Copyright (C) 2001-2018 Zabbix SIA
+** Copyright (C) 2001-2021 Zabbix SIA
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
@@ -70,6 +70,9 @@ class CApiInputValidator {
 	 */
 	private static function validateData($rule, &$data, $path, &$error, array $parent_data = null) {
 		switch ($rule['type']) {
+			case API_CALC_FORMULA:
+				return self::validateCalcFormula($rule, $data, $path, $error);
+
 			case API_COLOR:
 				return self::validateColor($rule, $data, $path, $error);
 
@@ -91,6 +94,18 @@ class CApiInputValidator {
 			case API_INTS32:
 				return self::validateInts32($rule, $data, $path, $error);
 
+			case API_UINT64:
+				return self::validateUInt64($rule, $data, $path, $error);
+
+			case API_UINTS64:
+				return self::validateUInts64($rule, $data, $path, $error);
+
+			case API_FLOAT:
+				return self::validateFloat($rule, $data, $path, $error);
+
+			case API_FLOATS:
+				return self::validateFloats($rule, $data, $path, $error);
+
 			case API_ID:
 				return self::validateId($rule, $data, $path, $error);
 
@@ -106,6 +121,12 @@ class CApiInputValidator {
 			case API_OUTPUT:
 				return self::validateOutput($rule, $data, $path, $error);
 
+			case API_PSK:
+				return self::validatePSK($rule, $data, $path, $error);
+
+			case API_SORTORDER:
+				return self::validateSortOrder($rule, $data, $path, $error);
+
 			case API_IDS:
 				return self::validateIds($rule, $data, $path, $error);
 
@@ -121,12 +142,15 @@ class CApiInputValidator {
 			case API_NUMERIC:
 				return self::validateNumeric($rule, $data, $path, $error);
 
-			case API_SCRIPT_NAME:
-				return self::validateScriptName($rule, $data, $path, $error);
+			case API_SCRIPT_MENU_PATH:
+				return self::validateScriptMenuPath($rule, $data, $path, $error);
 
 			case API_USER_MACRO:
 				return self::validateUserMacro($rule, $data, $path, $error);
 
+			case API_LLD_MACRO:
+				return self::validateLLDMacro($rule, $data, $path, $error);
+
 			case API_RANGE_TIME:
 				return self::validateRangeTime($rule, $data, $path, $error);
 
@@ -147,6 +171,36 @@ class CApiInputValidator {
 
 			case API_URL:
 				return self::validateUrl($rule, $data, $path, $error);
+
+			case API_IP:
+				return self::validateIp($rule, $data, $path, $error);
+
+			case API_DNS:
+				return self::validateDns($rule, $data, $path, $error);
+
+			case API_PORT:
+				return self::validatePort($rule, $data, $path, $error);
+
+			case API_TRIGGER_EXPRESSION:
+				return self::validateTriggerExpression($rule, $data, $path, $error);
+
+			case API_EVENT_NAME:
+				return self::validateEventName($rule, $data, $path, $error);
+
+			case API_JSONRPC_PARAMS:
+				return self::validateJsonRpcParams($rule, $data, $path, $error);
+
+			case API_JSONRPC_ID:
+				return self::validateJsonRpcId($rule, $data, $path, $error);
+
+			case API_DATE:
+				return self::validateDate($rule, $data, $path, $error);
+
+			case API_NUMERIC_RANGES:
+				return self::validateNumericRanges($rule, $data, $path, $error);
+
+			case API_UUID:
+				return self::validateUuid($rule, $data, $path, $error);
 		}
 
 		// This message can be untranslated because warn about incorrect validation rules at a development stage.
@@ -167,19 +221,27 @@ class CApiInputValidator {
 	 */
 	private static function validateDataUniqueness($rule, &$data, $path, &$error) {
 		switch ($rule['type']) {
+			case API_CALC_FORMULA:
 			case API_COLOR:
 			case API_MULTIPLE:
 			case API_STRING_UTF8:
 			case API_INT32:
+			case API_UINT64:
+			case API_UINTS64:
+			case API_FLOAT:
+			case API_FLOATS:
 			case API_ID:
 			case API_BOOLEAN:
 			case API_FLAG:
 			case API_OUTPUT:
+			case API_PSK:
+			case API_SORTORDER:
 			case API_HG_NAME:
 			case API_H_NAME:
 			case API_NUMERIC:
-			case API_SCRIPT_NAME:
+			case API_SCRIPT_MENU_PATH:
 			case API_USER_MACRO:
+			case API_LLD_MACRO:
 			case API_RANGE_TIME:
 			case API_TIME_PERIOD:
 			case API_TIME_UNIT:
@@ -187,6 +249,16 @@ class CApiInputValidator {
 			case API_HTTP_POST:
 			case API_VARIABLE_NAME:
 			case API_URL:
+			case API_IP:
+			case API_DNS:
+			case API_PORT:
+			case API_TRIGGER_EXPRESSION:
+			case API_EVENT_NAME:
+			case API_JSONRPC_PARAMS:
+			case API_JSONRPC_ID:
+			case API_DATE:
+			case API_NUMERIC_RANGES:
+			case API_UUID:
 				return true;
 
 			case API_OBJECT:
@@ -244,6 +316,57 @@ class CApiInputValidator {
 		return true;
 	}
 
+	/**
+	 * Calculated item formula validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_ALLOW_LLD_MACRO
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateCalcFormula($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8(API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		$expression_parser = new CExpressionParser([
+			'usermacros' => true,
+			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO),
+			'calculated' => true,
+			'host_macro' => true,
+			'empty_host' => true
+		]);
+
+		if ($expression_parser->parse($data) != CParser::PARSE_SUCCESS) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, $expression_parser->getError());
+			return false;
+		}
+
+		$expression_validator = new CExpressionValidator([
+			'usermacros' => true,
+			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO),
+			'calculated' => true
+		]);
+
+		if (!$expression_validator->validate($expression_parser->getResult()->getTokens())) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, $expression_validator->getError());
+			return false;
+		}
+
+		return true;
+	}
+
 	/**
 	 * Color validator.
 	 *
@@ -413,12 +536,12 @@ class CApiInputValidator {
 			return true;
 		}
 
-		if ((!is_int($data) && !is_string($data)) || 1 != preg_match('/^\-?[0-9]+$/', strval($data))) {
-			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is expected'));
+		if ((!is_int($data) && !is_string($data)) || !preg_match('/^'.ZBX_PREG_INT.'$/', strval($data))) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an integer is expected'));
 			return false;
 		}
 
-		if (bccomp($data, ZBX_MIN_INT32) < 0 || bccomp($data, ZBX_MAX_INT32) > 0) {
+		if ($data < ZBX_MIN_INT32 || $data > ZBX_MAX_INT32) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
 			return false;
 		}
@@ -434,6 +557,187 @@ class CApiInputValidator {
 		return true;
 	}
 
+	/**
+	 * Unsigned integers validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_ALLOW_NULL
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateUInt64($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (($flags & API_ALLOW_NULL) && $data === null) {
+			return true;
+		}
+
+		if (!is_scalar($data) || is_bool($data) || is_double($data) || !ctype_digit(strval($data))) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an unsigned integer is expected'));
+			return false;
+		}
+
+		if (bccomp($data, ZBX_MAX_UINT64) > 0) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
+			return false;
+		}
+
+		$data = (string) $data;
+
+		if ($data[0] === '0') {
+			$data = ltrim($data, '0');
+
+			if ($data === '') {
+				$data = '0';
+			}
+		}
+
+		return true;
+	}
+
+	/**
+	 * Array of unsigned integers validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_NULL, API_NORMALIZE
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateUInts64($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (($flags & API_ALLOW_NULL) && $data === null) {
+			return true;
+		}
+
+		if (($flags & API_NORMALIZE) && self::validateUInt64([], $data, '', $e)) {
+			$data = [$data];
+		}
+		unset($e);
+
+		if (!is_array($data)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an array is expected'));
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) && !$data) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('cannot be empty'));
+			return false;
+		}
+
+		$data = array_values($data);
+		$rules = ['type' => API_UINT64];
+
+		foreach ($data as $index => &$value) {
+			$subpath = ($path === '/' ? $path : $path.'/').($index + 1);
+			if (!self::validateData($rules, $value, $subpath, $error)) {
+				return false;
+			}
+		}
+		unset($value);
+
+		return true;
+	}
+
+	/**
+	 * Floating point number validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_ALLOW_NULL
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateFloat($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (($flags & API_ALLOW_NULL) && $data === null) {
+			return true;
+		}
+
+		if (is_int($data) || is_float($data)) {
+			$value = (float) $data;
+		}
+		elseif (is_string($data)) {
+			$number_parser = new CNumberParser();
+
+			if ($number_parser->parse($data) == CParser::PARSE_SUCCESS) {
+				$value = (float) $number_parser->getMatch();
+			}
+			else {
+				$value = NAN;
+			}
+		}
+		else {
+			$value = NAN;
+		}
+
+		if (is_nan($value)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a floating point value is expected'));
+
+			return false;
+		}
+
+		$data = $value;
+
+		return true;
+	}
+
+	/**
+	 * Array of floating point numbers validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_NULL, API_NORMALIZE
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateFloats($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (($flags & API_ALLOW_NULL) && $data === null) {
+			return true;
+		}
+
+		if (($flags & API_NORMALIZE) && self::validateFloat([], $data, '', $e)) {
+			$data = [$data];
+		}
+		unset($e);
+
+		if (!is_array($data)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an array is expected'));
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) && !$data) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('cannot be empty'));
+			return false;
+		}
+
+		$data = array_values($data);
+		$rules = ['type' => API_FLOAT];
+
+		foreach ($data as $index => &$value) {
+			$subpath = ($path === '/' ? $path : $path.'/').($index + 1);
+			if (!self::validateData($rules, $value, $subpath, $error)) {
+				return false;
+			}
+		}
+		unset($value);
+
+		return true;
+	}
+
 	private static function Int32In($data, $in) {
 		$valid = false;
 
@@ -485,7 +789,7 @@ class CApiInputValidator {
 	 *
 	 * @param array  $rule
 	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_NULL, API_NORMALIZE
-	 * @param string $rule['in']      (optional) a comma-delimited character string, for example: 'xml,json'
+	 * @param string $rule['in']      (optional) a comma-delimited character string, for example: '0,60:900'
 	 * @param mixed  $data
 	 * @param string $path
 	 * @param string $error
@@ -560,7 +864,7 @@ class CApiInputValidator {
 			return false;
 		}
 
-		if (bccomp($data, '9223372036854775807') > 0) {
+		if (bccomp($data, ZBX_DB_MAX_ID) > 0) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
 			return false;
 		}
@@ -582,6 +886,7 @@ class CApiInputValidator {
 	 * Boolean validator.
 	 *
 	 * @param array  $rule
+	 * @param int    $rule['flags']  (optional) API_ALLOW_NULL
 	 * @param mixed  $data
 	 * @param string $path
 	 * @param string $error
@@ -589,6 +894,12 @@ class CApiInputValidator {
 	 * @return bool
 	 */
 	private static function validateBoolean($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (($flags & API_ALLOW_NULL) && $data === null) {
+			return true;
+		}
+
 		if (!is_bool($data)) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a boolean is expected'));
 			return false;
@@ -690,7 +1001,7 @@ class CApiInputValidator {
 	}
 
 	/**
-	 * APPI output validator.
+	 * API output validator.
 	 *
 	 * @param array  $rule
 	 * @param int    $rule['flags']   (optional) API_ALLOW_COUNT, API_ALLOW_NULL
@@ -729,6 +1040,92 @@ class CApiInputValidator {
 		return false;
 	}
 
+	/**
+	 * PSK key validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validatePSK($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		$mb_len = mb_strlen($data);
+
+		if ($mb_len != 0 && $mb_len < PSK_MIN_LEN) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _s('minimum length is %1$s characters', PSK_MIN_LEN));
+			return false;
+		}
+
+		if (preg_match('/^([0-9a-f]{2})*$/i', $data) !== 1) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path,
+				_('an even number of hexadecimal characters is expected')
+			);
+			return false;
+		}
+
+		if (array_key_exists('length', $rule) && $mb_len > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * API sort order validator.
+	 *
+	 * @param array  $rule
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateSortOrder($rule, &$data, $path, &$error) {
+		$in = ZBX_SORT_UP.','.ZBX_SORT_DOWN;
+
+		if (self::validateStringUtf8(['in' => $in], $data, $path, $e)) {
+			return true;
+		}
+
+		if (is_string($data)) {
+			$error = $e;
+			return false;
+		}
+		unset($e);
+
+		if (!is_array($data)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an array or a character string is expected'));
+			return false;
+		}
+
+		$data = array_values($data);
+		$rules = [
+			'type' => API_STRING_UTF8,
+			'in' => $in
+		];
+
+		foreach ($data as $index => &$value) {
+			$subpath = ($path === '/' ? $path : $path.'/').($index + 1);
+			if (!self::validateData($rules, $value, $subpath, $error)) {
+				return false;
+			}
+		}
+		unset($value);
+
+		return true;
+	}
+
 	/**
 	 * Array of ids validator.
 	 *
@@ -779,8 +1176,9 @@ class CApiInputValidator {
 	 * Array of objects validator.
 	 *
 	 * @param array  $rule
-	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_NULL, API_NORMALIZE
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_NULL, API_NORMALIZE, API_PRESERVE_KEYS
 	 * @param array  $rule['fields']
+	 * @param int    $rule['length']  (optional)
 	 * @param mixed  $data
 	 * @param string $path
 	 * @param string $error
@@ -804,6 +1202,11 @@ class CApiInputValidator {
 			return false;
 		}
 
+		if (array_key_exists('length', $rule) && count($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
 		if (($flags & API_NORMALIZE) && $data) {
 			reset($data);
 
@@ -812,7 +1215,9 @@ class CApiInputValidator {
 			}
 		}
 
-		$data = array_values($data);
+		if (!($flags & API_PRESERVE_KEYS)) {
+			$data = array_values($data);
+		}
 
 		foreach ($data as $index => &$value) {
 			$subpath = ($path === '/' ? $path : $path.'/').($index + 1);
@@ -927,6 +1332,8 @@ class CApiInputValidator {
 	 * @return bool
 	 */
 	private static function validateNumeric($rule, &$data, $path, &$error) {
+		global $DB;
+
 		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
 
 		if (is_int($data)) {
@@ -939,6 +1346,7 @@ class CApiInputValidator {
 
 		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+
 			return false;
 		}
 
@@ -946,30 +1354,50 @@ class CApiInputValidator {
 			return true;
 		}
 
-		$pattern = '/^(-?)0*(0|[1-9][0-9]*)(\.?[0-9]+)?(['.ZBX_BYTE_SUFFIXES.ZBX_TIME_SUFFIXES.'])?$/';
+		$number_parser = new CNumberParser(['with_suffix' => true]);
 
-		if (1 != preg_match($pattern, strval($data))) {
+		if ($number_parser->parse($data) != CParser::PARSE_SUCCESS) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is expected'));
+
 			return false;
 		}
 
-		$value = convertFunctionValue($data, ZBX_UNITS_ROUNDOFF_LOWER_LIMIT);
+		$value = $number_parser->calcValue();
 
-		if (bccomp($value, ZBX_MIN_INT64) < 0 || bccomp($value, ZBX_MAX_INT64) > 0) {
-			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
-			return false;
+		if ($DB['DOUBLE_IEEE754']) {
+			if (abs($value) > ZBX_FLOAT_MAX) {
+				$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
+
+				return false;
+			}
+		}
+		else {
+			if (abs($value) >= 1E+16) {
+				$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
+
+				return false;
+			}
+			elseif ($value != round($value, 4)) {
+				$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number has too many fractional digits'));
+
+				return false;
+			}
 		}
 
-		// Trim leading zeroes.
-		$data = preg_replace($pattern, '${1}${2}${3}${4}', (string) $data);
+		// Remove leading zeros.
+		$data = preg_replace('/^(-)?(0+)?(\d.*)$/', '${1}${3}', $data);
+
+		// Add leading zero.
+		$data = preg_replace('/^(-)?(\..*)$/', '${1}0${2}', $data);
 
 		return true;
 	}
 
 	/**
-	 * Global script name validator.
+	 * Global script menu_path validator.
 	 *
 	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional)
 	 * @param int    $rule['length']  (optional)
 	 * @param mixed  $data
 	 * @param string $path
@@ -977,8 +1405,17 @@ class CApiInputValidator {
 	 *
 	 * @return bool
 	 */
-	private static function validateScriptName($rule, &$data, $path, &$error) {
-		if (self::checkStringUtf8(API_NOT_EMPTY, $data, $path, $error) === false) {
+	private static function validateScriptMenuPath($rule, &$data, $path, &$error) {
+		// Having only a root folder is the same as being empty. Temporary modify data to check if it is actually empty.
+		$tmp_data = $data;
+
+		if ($tmp_data === '/') {
+			$tmp_data = '';
+		}
+
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags, $tmp_data, $path, $error) === false) {
 			return false;
 		}
 
@@ -987,13 +1424,21 @@ class CApiInputValidator {
 			return false;
 		}
 
+		// If empty is allowed there is only root folder, return early.
+		if ($data === '/') {
+			return true;
+		}
+
 		$folders = splitPath($data);
 		$folders = array_map('trim', $folders);
+		$count = count($folders);
 
 		// folder1/{empty}/name or folder1/folder2/{empty}
-		foreach ($folders as $folder) {
-			if ($folder === '') {
-				$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('directory or script name cannot be empty'));
+		foreach ($folders as $num => $folder) {
+			// Allow the trailing slash.
+			if ($folder === '' && $num != ($count - 1) && $num != 0) {
+				$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('directory cannot be empty'));
+
 				return false;
 			}
 		}
@@ -1030,6 +1475,35 @@ class CApiInputValidator {
 		return true;
 	}
 
+	/**
+	 * LLD macro validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateLLDMacro($rule, &$data, $path, &$error) {
+		if (self::checkStringUtf8(API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		if ((new CLLDMacroParser())->parse($data) != CParser::PARSE_SUCCESS) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a low-level discovery macro is expected'));
+			return false;
+		}
+
+		return true;
+	}
+
 	/**
 	 * Time period validator like "1-7,00:00-24:00".
 	 *
@@ -1125,7 +1599,7 @@ class CApiInputValidator {
 			return true;
 		}
 
-		if (false === @preg_match('/'.$data.'/', '')) {
+		if (@preg_match('/'.str_replace('/', '\/', $data).'/', '') === false) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('invalid regular expression'));
 			return false;
 		}
@@ -1137,8 +1611,10 @@ class CApiInputValidator {
 	 * Time unit validator like "10", "20s", "30m", "4h", "{$TIME}" etc.
 	 *
 	 * @param array  $rule
-	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_USER_MACRO, API_ALLOW_LLD_MACRO
-	 * @param int    $rule['in']      (optional)
+	 * @param int    $rule['length'] (optional)
+	 * @param int    $rule['flags']  (optional) API_NOT_EMPTY, API_ALLOW_USER_MACRO, API_ALLOW_LLD_MACRO,
+	 *                                          API_TIME_UNIT_WITH_YEAR
+	 * @param int    $rule['in']     (optional)
 	 * @param mixed  $data
 	 * @param string $path
 	 * @param string $error
@@ -1160,6 +1636,11 @@ class CApiInputValidator {
 			return false;
 		}
 
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
 		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
 			return true;
 		}
@@ -1167,7 +1648,8 @@ class CApiInputValidator {
 		$simple_interval_parser = new CSimpleIntervalParser([
 			'usermacros' => ($flags & API_ALLOW_USER_MACRO),
 			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO),
-			'negative' => true
+			'negative' => true,
+			'with_year' => ($flags & API_TIME_UNIT_WITH_YEAR)
 		]);
 
 		if ($simple_interval_parser->parse($data) != CParser::PARSE_SUCCESS) {
@@ -1179,9 +1661,9 @@ class CApiInputValidator {
 			return true;
 		}
 
-		$seconds = timeUnitToSeconds($data);
+		$seconds = timeUnitToSeconds($data, ($flags & API_TIME_UNIT_WITH_YEAR));
 
-		if (bccomp(ZBX_MIN_INT32, $seconds) > 0 || bccomp($seconds, ZBX_MAX_INT32) > 0) {
+		if ($seconds < ZBX_MIN_INT32 || $seconds > ZBX_MAX_INT32) {
 			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is too large'));
 			return false;
 		}
@@ -1374,7 +1856,7 @@ class CApiInputValidator {
 	 *
 	 * @param array  $rule
 	 * @param int    $rule['length']  (optional)
-	 * @param int    $rule['flags']   (optional) API_ALLOW_USER_MACRO, API_NOT_EMPTY
+	 * @param int    $rule['flags']   (optional) API_ALLOW_USER_MACRO, API_ALLOW_EVENT_TAGS_MACRO, API_NOT_EMPTY
 	 * @param mixed  $data
 	 * @param string $path
 	 * @param string $error
@@ -1393,11 +1875,403 @@ class CApiInputValidator {
 			return false;
 		}
 
-		if ($data !== '' && CHtmlUrlValidator::validate($data, ($flags & API_ALLOW_USER_MACRO)) === false) {
-			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('unacceptible URL'));
+		$options = [
+			'allow_user_macro' => (bool) ($flags & API_ALLOW_USER_MACRO),
+			'allow_event_tags_macro' => (bool) ($flags & API_ALLOW_EVENT_TAGS_MACRO)
+		];
+
+		if ($data !== '' && CHtmlUrlValidator::validate($data, $options) === false) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('unacceptable URL'));
 			return false;
 		}
 
 		return true;
 	}
+
+	/**
+	 * IP address validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_USER_MACRO, API_ALLOW_LLD_MACRO,
+	 *                                API_ALLOW_MACRO
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateIp($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
+			return true;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		$ip_parser = new CIPParser([
+			'v6' => ZBX_HAVE_IPV6,
+			'usermacros' => ($flags & API_ALLOW_USER_MACRO),
+			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO),
+			'macros' => ($flags & API_ALLOW_MACRO)
+		]);
+
+		if ($ip_parser->parse($data) != CParser::PARSE_SUCCESS) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an IP address is expected'));
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * DNS name validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_USER_MACRO, API_ALLOW_LLD_MACRO,
+	 *                                API_ALLOW_MACRO
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateDns($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
+			return true;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		$dns_parser = new CDnsParser([
+			'usermacros' => ($flags & API_ALLOW_USER_MACRO),
+			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO),
+			'macros' => ($flags & API_ALLOW_MACRO)
+		]);
+
+		if ($dns_parser->parse($data) != CParser::PARSE_SUCCESS) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a DNS name is expected'));
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * Port number validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY, API_ALLOW_USER_MACRO, API_ALLOW_LLD_MACRO
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validatePort($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (!is_int($data) && !is_string($data)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a number is expected'));
+			return false;
+		}
+
+		$data = (string) $data;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
+			return true;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		if ($flags & API_ALLOW_USER_MACRO) {
+			$user_macro_parser = new CUserMacroParser();
+
+			if ($user_macro_parser->parse($data) == CParser::PARSE_SUCCESS) {
+				return true;
+			}
+		}
+
+		if ($flags & API_ALLOW_LLD_MACRO) {
+			$lld_macro_parser = new CLLDMacroParser();
+
+			if ($lld_macro_parser->parse($data) == CParser::PARSE_SUCCESS) {
+				return true;
+			}
+		}
+
+		if (!self::validateInt32(['in' => ZBX_MIN_PORT_NUMBER.':'.ZBX_MAX_PORT_NUMBER], $data, $path, $error)) {
+			return false;
+		}
+
+		$data = (string) $data;
+
+		return true;
+	}
+
+	/**
+	 * Trigger expression validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['length']  (optional)
+	 * @param int    $rule['flags']   (optional) API_ALLOW_LLD_MACRO, API_NOT_EMPTY
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateTriggerExpression($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
+			return true;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		$expression_parser = new CExpressionParser([
+			'usermacros' => true,
+			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO)
+		]);
+
+		if ($expression_parser->parse($data) != CParser::PARSE_SUCCESS) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, $expression_parser->getError());
+			return false;
+		}
+
+		$expression_validator = new CExpressionValidator([
+			'usermacros' => true,
+			'lldmacros' => ($flags & API_ALLOW_LLD_MACRO)
+		]);
+
+		if (!$expression_validator->validate($expression_parser->getResult()->getTokens())) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, $expression_validator->getError());
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * Event name validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateEventName($rule, &$data, $path, &$error) {
+		if (self::checkStringUtf8(0, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		$eventname_validator = new CEventNameValidator();
+
+		if (!$eventname_validator->validate($data)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, $eventname_validator->getError());
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * JSON RPC parameters validator. Parameters MUST contain an array or object value.
+	 *
+	 * @param array  $rule
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateJsonRpcParams($rule, &$data, $path, &$error) {
+		if (is_array($data)) {
+			return true;
+		}
+
+		$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('an array or object is expected'));
+
+		return false;
+	}
+
+	/**
+	 * JSON RPC identifier validator. This identifier MUST contain a String, Number, or NULL value.
+	 *
+	 * @param array  $rule
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateJsonRpcId($rule, &$data, $path, &$error) {
+		if (is_string($data) || is_int($data) || is_float($data) || $data === null) {
+			return true;
+		}
+
+		$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a string, number or null value is expected'));
+
+		return false;
+	}
+
+	/**
+	 * Date validator in YYYY-MM-DD format.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']  (optional) API_NOT_EMPTY
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateDate(array $rule, &$data, string $path, string &$error): bool {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
+			return true;
+		}
+
+		[$year, $month, $day] = sscanf($data, '%d-%d-%d');
+
+		if (!checkdate($month, $day, $year)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('a date in YYYY-MM-DD format is expected'));
+			return false;
+		}
+
+		if (!validateDateInterval($year, $month, $day)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path,
+				_s('value must be between "%1$s" and "%2$s"', '1970-01-01', '2038-01-18')
+			);
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * Validate numeric ranges. Multiple ranges separated by comma character.
+	 * Example:
+	 *   10-20,-20--10,-5-0,0.5-0.7,-20--10,-20.20--20.10
+	 *   30,-10,0.7,-0.5
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']   (optional) API_NOT_EMPTY
+	 * @param int    $rule['length']  (optional)
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateNumericRanges($rule, &$data, $path, &$error) {
+		$flags = array_key_exists('flags', $rule) ? $rule['flags'] : 0x00;
+
+		if (self::checkStringUtf8($flags & API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (($flags & API_NOT_EMPTY) == 0 && $data === '') {
+			return true;
+		}
+
+		if (array_key_exists('length', $rule) && mb_strlen($data) > $rule['length']) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('value is too long'));
+			return false;
+		}
+
+		$parser = new CRangesParser(['with_minus' => true, 'with_float' => true, 'with_suffix' => true]);
+
+		if ($parser->parse($data) != CParser::PARSE_SUCCESS) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('invalid range expression'));
+
+			return false;
+		}
+
+		return true;
+	}
+
+	/**
+	 * UUIDv4 validator.
+	 *
+	 * @param array  $rule
+	 * @param int    $rule['flags']  (optional) API_NOT_EMPTY
+	 * @param mixed  $data
+	 * @param string $path
+	 * @param string $error
+	 *
+	 * @return bool
+	 */
+	private static function validateUuid(array $rule, &$data, string $path, string &$error): bool {
+		if (self::checkStringUtf8(API_NOT_EMPTY, $data, $path, $error) === false) {
+			return false;
+		}
+
+		if (mb_strlen($data) != 32) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _s('must be %1$s characters long', 32));
+			return false;
+		}
+
+		if (!ctype_xdigit($data)) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('UUIDv4 is expected'));
+			return false;
+		}
+
+		$binary = hex2bin($data);
+		if ((ord($binary[6]) & 0xf0) != 0x40 || (ord($binary[8]) & 0xc0) != 0x80) {
+			$error = _s('Invalid parameter "%1$s": %2$s.', $path, _('UUIDv4 is expected'));
+			return false;
+		}
+
+		$data = strtolower($data);
+
+		return true;
+	}
 }
